// ddlists.cpp : Defines the entry point for the console application.
//

#include "stdafx.h"
#include "dlists.h"
#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>

// compare two data elements
bool
is_equal(Data one,Data two)
{
      return( (one.x == two.x) && (one.y == two.y) );
} // is_equal

// add an element to the end of the list
int
add_end(List_data* & l, Data my_d)
{
   List_data *lptr = l,*my_l;
   if( (my_l = new List_data) == NULL) return -1;

   my_l->next = NULL;  // since this will be the new last element
   my_l->d = my_d;

   // if there are no nodes on the list, then the new one is
   // both the first and last node, so l (remember it is a
   // reference parameter) should point at it.
   if(lptr == NULL) {
      my_l->prev = l; // and it should point at l
      l = my_l;
   }
   else {
      // move lptr until it points at the current last node
      while(lptr->next != NULL) lptr = lptr->next;
      lptr->next = my_l; // point it at our new node
      my_l->prev = lptr; // and it should point at l
   }
   return 0;
} // add_element

// search the list node by node until we find the one
// that contains my_d. Return NULL if we don't find it
List_data *
find_element(List_data *l, Data my_d)
{
   List_data * lptr = l;
   while(lptr != NULL) {
      if( is_equal(lptr->d,my_d) ){
         return(lptr); // found it
      }
      lptr = lptr->next;
   } // while
   return(NULL);
} // find_element

// add a new node to the front of the list.
int
add_front(List_data * & l, Data my_d)
{
   List_data *lp;
   if( (lp = new List_data) == NULL) return -1;
   lp->d = my_d;
   // since this will be the new first one,
   // it should point to the current first one
   lp->next = l;
   lp->prev = NULL;  // since this is  the first one
   l = lp; // make the new one the first one.
   return 0;
} // add_front

// search the list for the target data and add a new node after
// that one.
int
insert(List_data* l, Data target,Data my_d)
{
   List_data *lp,*lptr;
   
   if((lp = new List_data) == NULL) return -1;
   lp->d = my_d;
   // use the find_element() function above to locate target
   // in the list, if it is there
   lptr = find_element(l, target);
   if(lptr == NULL)  // not there
      return(-1);
   else {
      // the new one should point to the one
      // after the one we found (lptr->next)
      // and also to the target one
      lp->prev = lptr;
      lp->next = lptr->next;
      // the one after the target should point to the new one
      lptr->next->prev = lp;
      // the one we found should point to the new one
      lptr->next = lp;
      // so the new one is after the one we found
   }
   return 0;   
} // insert

// same as above, but takes uses the fact we are adding after the target.
// The one find_element() returns (lptr) can be seen as the front of a list 
// that starts with the one that comes after it.
// So we want to add to the front of that list.
// Which is what add_front() does.
int
insert2(List_data* &l, Data target,Data my_d)
{
   List_data *lptr;
   
   lptr = find_element(l, target);
   if(lptr == NULL)
      return(-1);  // didn't find it
   else
      add_front(lptr->next,my_d);
   return 0;   
} // insert

// walk the list, printing each data element
void
print_list(List_data *l)
{
   List_data * lptr = l;
   
   cout << "list: ";
   while(lptr != NULL) {
      cout << "(" <<  lptr->d.x << "," << lptr->d.y << ") ";
      lptr = lptr->next;
   } // while
   cout << endl;
} // print_list

// remove the node that contains the data my_d
int
delete1(List_data *l, Data my_d)
{
   List_data *current;
   // if the first one is the one we want, delete it quick
   if( is_equal(l->d,my_d) ){
      l = l->next;
      l->prev = NULL;
      return 0;
   }

   // if we find it, make the one that comes before it
   // point to  the one that comes after. This deletes the node
   // from the list
   current = l;
   while(current != NULL) {
      if( is_equal(current->d,my_d) ){ 
	 // make the one before the target point to the one after the
	 // target. And the one after should point to the one before
	 current->prev->next = current->next;
	 current->next->prev = current->prev;
      }
      // move the two pointers to the next nodes
      current = current->next;
   } // while

   return -1;  // didn't find it
} // delete1
