// slists.cpp : a simple linked list using structures

#include "stdafx.h"
#include <conio.h>
#include <iostream>
using namespace std;
const int STRSIZE = 64;

// each instance of this struct represents one character in the string
struct cnode {
        char c;  // data for this node
        cnode * next;  // pointer to the next node
}; // cnode

// this reads in char by char, counting how many have been entered
// if the number exceeds the length passed in, the reading stops
// This assumes that the array pointed to by str is long enough
void
get_string(char * str,int length,char * prompt)
{
    char c, *ptr = str;
    cout << prompt << " (limit " << length << ") " << endl;
    while(true) {
        c = _getche();   //reads an echoes one character
        if(c == '\r') break; // user hit enter key
        *ptr++ = c;        
        if((ptr - str) > (length -1 )) {
                cout << " no more characters can be entered" << endl;
                break;
        }
    } //while read characters
    *ptr = '\0';
    putchar('\n'); //cout << endl;
} // get_string

// add a data element to the front of the list. Returns true if OK
bool
add_to_front(cnode * & front,char data)
{
    cnode * new_node;

    if( (new_node = new cnode) == NULL) {
        cerr << " no memory for cnode \n";
        return false;
    }
    new_node->c = data;
    new_node->next=front;
    front = new_node;
    return true;
} // add_to_front

// add a data element to the back of the list. Returns true if OK
bool
add_to_end(cnode * & front,char data)
{
    cnode * new_node;
    cnode * ptr = front;

    if( (new_node = new cnode) == NULL) {
        cerr << " no memory for cnode \n";
        return false;
    }
    new_node->c = data;
    new_node->next = NULL;

    //if the list is empty then the last is the first, so put it
    //on the front of the list
    if(front == NULL) {
            cout << "add first one" << endl;
            front = new_node;
            return true;
    }

    // if not, we have to find the last element in the list
    // The last element is the one whose next pointer is NULL
    while(ptr->next != NULL) ptr = ptr->next;
    // now ptr points to the last element in the list
    // so make it point to the new one
    ptr->next = new_node;
    return true;
} // add_to_end

// put the characters from the cstring into a list
void
strtolist(cnode *& front,char * source)
{
        char * ptr = source;
        while(*ptr) {
//                cout << "adding char " << (char) *ptr << endl;
                if(!add_to_end(front,*ptr)) {
                        cerr << "adding char " << (char) *ptr << "failed" << endl;
                        return;
                }
                ptr++;
        } // while
} // strtolist

// print the characters in the string list and a new line at the end
void
print_list(cnode * front)
{
        cnode *ptr = front;  // always copy the pointer
        if(front == NULL) return;   // nothing to print
        while(ptr != NULL) {
                cout << ptr->c;
                ptr = ptr->next;
        } // while
        cout << endl;
} // print_list

// search the list node by node until we find the one
// that contains target. Return NULL if we don't find it
cnode *
find_element(cnode * front,char target)
{
   cnode * lptr = front;
   while(lptr != NULL) {
      if( lptr->c == target ){
         return(lptr); // found it
      }
      lptr = lptr->next;
   } // while
   return(NULL);
} // find_element

// search the list for the target data and add a new node after
// that one.
bool
add_after(cnode * front,char target,char newdata)
{
   cnode *lp,*lptr;
   
   // use the find_element() function above to locate target
   // in the list, if it is there
   lptr = find_element(front,target);
   if(lptr == NULL)  return(false); // not there
   if((lp = new cnode) == NULL) return false;   // make a new node
   lp->c = newdata; lp->next = NULL;
      
   // the new one should point to the one
   // after the one we found that contains the target (lptr->next)
   lp->next = lptr->next;
   // the one we found should point to the new one
   // so the new one is after the one we found
   lptr->next=lp;
   return true;   // that's all
} // add_after

// search the list for the target data and add a new node before
// that one.
// This is harder as we need to keep track of the target node and the
// node that comes before it so we can put the new node in between
bool
add_before(cnode *& front,char target,char newdata)
{
   cnode *lp,*targetptr,*beforeptr;
   
   if(front == NULL) return false; // nothing to do
   if((lp = new cnode) == NULL) return false;
   lp->c = newdata; lp->next = NULL;

   // Let's check the case where the target is the first node.
   // This is special because if it is, we have to change the front pointer
   // to point to the new node.
   if(front->c == target) {
		lp->next = front;  // make the new one point at the ones after front
		front = lp;  //  the new one is the new first one
		return true;
   }
   
   // first we find the target node, using the same
   // kind of code as in find_element() but with one
   // extra pointer to keep track of the previous node
   beforeptr = front; targetptr = front->next;
   while(targetptr != NULL) {
      if( targetptr->c == target ){    // found it
			  // we want the one that beforeptr points at to now point at
			  // the new one.
			  beforeptr->next = lp;
			  // and we want the new one to point at the target one
			  lp->next = targetptr;
			  return true;  // all done
      }
	  // keep looking
	  beforeptr = targetptr;
      targetptr = targetptr->next;
   } // while
   return false;    // didn't find it
} // add_before

// remove the node that contains the data my_d
bool
remove(cnode * front,char target)
{
   cnode *targetptr,*beforeptr;

   if(front == NULL) return false;  // nothing to delete

   // if the first one is the one we want, delete it quick
   if( front->c == target ){
      front = front->next;
      return true;
   }

   // if we find it, make the one that comes before it
   // point to  the one that comes after. This deletes the node
   // from the list
   beforeptr = front; targetptr = front->next;
   while(targetptr != NULL) {
      if( targetptr->c == target ){    // found it
			  // we want the one that beforeptr points at to now point at
			  // the new one.
			  // and we want the new one to point at the target one
			  beforeptr->next = targetptr->next;
			  delete targetptr;   // done with it
			  return true;  // all done
      }
	  // keep looking
	  beforeptr = targetptr;
      targetptr = targetptr->next;
   } // while
   return false;    // didn't find it
} // remove

int 
main(int argc, char* argv[])
{
    cnode * listfront = NULL;  // list starts out empty
	char target,newone,indata[STRSIZE];

    get_string(indata,STRSIZE, "Enter the string to be made into a list");
    strtolist(listfront,indata);
	cout << "The initial string version of " << indata << endl;
    print_list(listfront);

	cout << "Search for what character? ";
	target = _getche();  cout << endl;
	if(find_element(listfront,target) == NULL)
			cout << target << " wasn't there" << endl;
	else
			cout << target << " was there" << endl;

	cout << "Add what character after " << target << " ? ";
	newone = _getche();  cout << endl;
	if(add_after(listfront,target,newone) == NULL)
			cout << target << " couldn't be added" << endl;
	else
			cout << target << " was added" << endl;
	cout << "The string after adding " << newone << " is ";
    print_list(listfront);

	cout << "Add what character before " << target << " ? ";
	newone = _getche();  cout << endl;
	if(add_before(listfront,target,newone) == NULL)
			cout << target << " couldn't be added" << endl;
	else
			cout << target << " was added" << endl;
	cout << "The string after adding " << newone << " is ";
    print_list(listfront);

	cout << "delete what character? ";
	target = _getche();  cout << endl;
	if(remove(listfront,target) == NULL)
			cout << target << " couldn't be removed" << endl;
	else
			cout << target << " was removed" << endl;
	cout << "The string after removing " << target << " is ";
    print_list(listfront);
    return 0;
}
