// tree.cpp - code for the tree class
#include "stdafx.h"
#include <iostream>
#include <string>
#include "tree.h"
#include "except.h"
using namespace std;
extern ostream Os;

bool
tree::add_value(string str)
{
    if(str.length() == 0)
        throw new Err("add_value: bad parameter");
	Os << "\nAdding <" << str << "> to the tree"<< endl;
    return(add_valuer(root,str,0));
} // add_value

// this searches the tree for str and if it doesn't find
// it, it creates a node for it. It returns true if it
// either added a new node or found the value already in 
// the tree
bool
tree::add_valuer(node * &t,string str,int count)
{
    node * n;

    if(t == NULL) {
        if(debug) {
           for(int i=0; i< count; i++) Os << "  ";
		   Os << "adding to NULL tree" << endl;
		}
        if( (n = new node) == NULL) 
           throw new Mem("add_valuer: node");
        n->value=str;
        n->left = n->right = NULL;
        t = n;
        return(true);
     }
     else {
        int c = str.compare(t->value);
        if(debug) {
           for(int i=0; i< count; i++) Os << "  ";
		   Os << "root value is :" << t->value << ":\n";
		}
         if(c == 0)  {
             if(debug) {
                for(int i=0; i< count; i++) Os << "  ";
                Os << "already in the tree" << endl;
		     }
             return(true);
         }
         else
              if ( c > 0) {
        		 if(debug) {
           		    for(int i=0; i< count; i++) Os << "  ";
                    Os << "recursively adding right" << endl;
				 }
                 return(add_valuer(t->right,str,count+1));
              }
              else {
        		 if(debug) {
           		    for(int i=0; i< count; i++) Os << "  ";
                    Os << "recursively adding left" << endl;
				 }
                 return(add_valuer(t->left,str,count+1));
              }
     }  // t not NULL
} // add_valuer

// inorder tree print
void
tree::inprint()
{
    inprintr(root);
} // inprint

void
tree::inprintr(node * r)
{
     if(r == NULL) return;
     inprintr(r->left);
     Os << r->value << endl;
     inprintr(r->right);
} // inprintr

// preorder tree print
void
tree::preprint()
{
    preprintr(root);
} // preprint

void
tree::preprintr(node * r)
{
     if(r == NULL) return;
     Os << r->value << endl;
     preprintr(r->left);
     preprintr(r->right);
} // preprintr

// postorder tree print
void
tree::postprint()
{
    postprintr(root);
} // postprint

void
tree::postprintr(node * r)
{
     if(r == NULL) return;
     postprintr(r->left);
     postprintr(r->right);
     Os << r->value << endl;
} // postprintr

void
tree::treeprint()
{
    treeprintr(root,0);
} // treeprint

// this prints the tree T sideways showing it's structure. 
// L is a helping variable to keep a count of the indentation to print
void
tree::treeprintr(node * t, int l)
{
   if(t != NULL) {
      treeprintr(t->right,l+1);
      for(int i=0; i< l; i++) Os << "  ";
      Os << t->value << endl;
      treeprintr(t->left,l+1);
   }
} // treeprintr

bool
tree::find(string str)
{
    if(str.length() == 0)
       throw new Err("find: bad parameter");
    return(findr(root,str));
} // find

// this searches the tree for str 
// It returns true if it
// found the value and false otherwise
bool
tree::findr(node * t, string str)
{

    if(t == NULL) return(false);
        
    int c = str.compare(t->value);
    if(c == 0)  return(true);  // found it
    if ( c > 0) 
        return(findr(t->right,str));
    else 
         return(findr(t->left,str));
} // findr

// Binary Tree Deletion

bool
tree::remove(string val)
{
    if(val.length() == 0)
       throw new Err("remove: bad parameter");
    return(remover(val,root));
} // remove

void 
tree::del(node *& r, node * h)
{
   // r is passed by reference
   if(r == NULL) if(debug) Os << "in del, r is NULL" << endl;
   else
   	if(debug) Os << "in del: r->value is :" << r->value << ":" << endl;
   if(r->right != NULL)
       del(r->right,h);
   else {
       h->value=r->value;
       r = r->left;
   }
} // del

// T is passed by reference.
// Hunt through T for the node containing val and delete it.  
// Use a private method called del().
bool
tree::remover(string val,node *& T)
{
    node * hold;
    if(T == NULL) return false;
           
    if(debug) Os << "t->value is :" << T->value << ":"  << endl;
    if(val.compare(T->value) < 0 ) {
       if(debug) Os << "deleting down the left" << endl;
       return(remover(val,T->left));
    }
    
    if(val.compare(T->value) > 0) {
       if(debug) Os << "deleting down the right" << endl;
       return(remover(val,T->right));
    }
    // found it
    if(debug) Os << "Found it" << endl;
    hold = T;   // remember where we are
    if(hold->right == NULL) {
       if(debug) Os << "Moving up the left" << endl;
       T = hold->left;   // move left child pointer up
       return true;
    }
    
    if(hold->left == NULL) {
      if(debug) Os << "Moving up the right" << endl;
      T = hold->right;   // move right child pointer up
      return true;
    }
     
    if(debug) Os << "Calling del()" << endl;
    del(hold->left,hold);
    return true;
} // remover

bool
tree::remove2(string val)
{
    if(val.length() == 0)
       throw new Err("remove2: bad parameter");
    return(remover2(val,root));
} // remove2

void 
tree::del2(node *& r, node * h)
{
   // r is passed by reference
   if(r == NULL) if(debug) Os << "in del2, r is NULL" << endl;
   else
   	if(debug) Os << "in del2: r->value is :" << r->value << ":" << endl;
   if(r->left != NULL)
       del2(r->left,h);
   else {
       h->value=r->value;
       r = r->right;
   }
} // del2

// T is passed by reference.
// Hunt through T for the node containing val and delete it.  
// Use a private method called del().
bool
tree::remover2(string val,node *& T)
{
    node * hold;
    if(T == NULL) return false;
           
    if(debug) Os << "t->value is :" << T->value << ":"  << endl;
    if(val.compare(T->value) < 0 ) {
       if(debug) Os << "deleting down the left" << endl;
       return(remover2(val,T->left));
    }
    
    if(val.compare(T->value) > 0) {
       if(debug) Os << "deleting down the right" << endl;
       return(remover2(val,T->right));
    }
    // found it
    if(debug) Os << "Found it" << endl;
    hold = T;   // remember where we are
    if(hold->left == NULL) {
       if(debug) Os << "Moving up the left" << endl;
       T = hold->right;   // move left child pointer up
       return true;
    }
    
    if(hold->right == NULL) {
      if(debug) Os << "Moving up the right" << endl;
      T = hold->left;   // move right child pointer up
      return true;
    }
     
    if(debug) Os << "Calling del()" << endl;
    del2(hold->right,hold);
    return true;
} // remover

/*
This is an iterative version of the binary tree insertion algorithm.
This is most closely related to the search-insert routine in the 
recursive tree routines handout, Trees 1.
This particular implementation is from "Applied Data Structures Using Pascal",
Guy J. Hale, Richard J. Easton, D. C. Heath and Company,1987.
*/
void
tree::inserti(string val)
{
	insert2(root,val);
} // inserti

void
tree::insert2(node * &root,string val)
{
        node * newnode, *top;
        bool inserted = false;

        if( (newnode = new node) == NULL) 
           throw new Mem("inserti: node");
        newnode->value=val;
        newnode->right = newnode->left = NULL;
        top = root;
        if(root == NULL)
            root = newnode;
        else {
            while(!inserted) {
                if(val.compare(top->value) < 0)
                    if(top->left != NULL)
                        top = top->left;
                    else {
                        top->left = newnode;
                        inserted = true;
		    }
                else
                    if(top->right != NULL)
                        top = top->right;
                    else {
                        top->right = newnode;
                        inserted = true;
		    }
	    } // while
	} // root not NULL
} // insert2
